/*
 * OpenVINS: An Open Platform for Visual-Inertial Research
 * Copyright (C) 2018-2022 Patrick Geneva
 * Copyright (C) 2018-2022 Guoquan Huang
 * Copyright (C) 2018-2022 OpenVINS Contributors
 * Copyright (C) 2018-2019 Kevin Eckenhoff
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef OV_INIT_CERES_JPLQUATLOCAL_H
#define OV_INIT_CERES_JPLQUATLOCAL_H

#include <ceres/ceres.h>

namespace ov_init {

/**
 * @brief JPL quaternion CERES state parameterization
 */
class State_JPLQuatLocal : public ceres::LocalParameterization {
  public:
    /**
     * @brief State update function for a JPL quaternion representation.
     *
     * Implements update operation by left-multiplying the current
     * quaternion with a quaternion built from a small axis-angle perturbation.
     *
     * @f[
     * \bar{q}=norm\Big(\begin{bmatrix} 0.5*\mathbf{\theta_{dx}} \\ 1 \end{bmatrix}\Big) \hat{\bar{q}}
     * @f]
     */
    bool Plus(const double *x, const double *delta, double *x_plus_delta) const override;

    /**
     * @brief Computes the jacobian in respect to the local parameterization
     *
     * This essentially "tricks" ceres.
     * Instead of doing what ceres wants:
     * dr/dlocal= dr/dglobal * dglobal/dlocal
     *
     * We instead directly do:
     * dr/dlocal= [ dr/dlocal, 0] * [I; 0]= dr/dlocal.
     * Therefore we here define dglobal/dlocal= [I; 0]
     */
    bool ComputeJacobian(const double *x, double *jacobian) const override;

    int GlobalSize() const override { return 4; };

    int LocalSize() const override { return 3; };
};

} // namespace ov_init

#endif // OV_INIT_CERES_JPLQUATLOCAL_H